import logging
import re
from pathlib import Path
from typing import Dict, List, Tuple

from khoj.database.models import Entry as DbEntry
from khoj.database.models import KhojUser
from khoj.processor.content.org_mode import orgnode
from khoj.processor.content.org_mode.orgnode import Orgnode
from khoj.processor.content.text_to_entries import TextToEntries
from khoj.utils import state
from khoj.utils.helpers import timer
from khoj.utils.rawconfig import Entry

logger = logging.getLogger(__name__)


class OrgToEntries(TextToEntries):
    def __init__(self):
        super().__init__()

    # Define Functions
    def process(self, files: dict[str, str], user: KhojUser, regenerate: bool = False) -> Tuple[int, int]:
        deletion_file_names = set([file for file in files if files[file] == ""])
        files_to_process = set(files) - deletion_file_names
        files = {file: files[file] for file in files_to_process}

        # Extract Entries from specified Org files
        max_tokens = 256
        with timer("Extract entries from specified Org files", logger):
            file_to_text_map, current_entries = self.extract_org_entries(files, max_tokens=max_tokens)

        with timer("Split entries by max token size supported by model", logger):
            current_entries = self.split_entries_by_max_tokens(current_entries, max_tokens=max_tokens)

        # Identify, mark and merge any new entries with previous entries
        with timer("Identify new or updated entries", logger):
            num_new_embeddings, num_deleted_embeddings = self.update_embeddings(
                user,
                current_entries,
                DbEntry.EntryType.ORG,
                DbEntry.EntrySource.COMPUTER,
                "compiled",
                logger,
                deletion_file_names,
                regenerate=regenerate,
                file_to_text_map=file_to_text_map,
            )

        return num_new_embeddings, num_deleted_embeddings

    @staticmethod
    def extract_org_entries(
        org_files: dict[str, str], index_heading_entries: bool = False, max_tokens=256
    ) -> Tuple[Dict, List[Entry]]:
        "Extract entries from specified Org files"
        file_to_text_map, entries, entry_to_file_map = OrgToEntries.extract_org_nodes(org_files, max_tokens)
        return file_to_text_map, OrgToEntries.convert_org_nodes_to_entries(
            entries, entry_to_file_map, index_heading_entries
        )

    @staticmethod
    def extract_org_nodes(
        org_files: dict[str, str], max_tokens
    ) -> Tuple[Dict, List[List[Orgnode]], Dict[Orgnode, str]]:
        "Extract org nodes from specified org files"
        entries: List[List[Orgnode]] = []
        entry_to_file_map: List[Tuple[Orgnode, str]] = []
        file_to_text_map = {}
        for org_file in org_files:
            try:
                org_content = org_files[org_file]
                entries, entry_to_file_map = OrgToEntries.process_single_org_file(
                    org_content, org_file, entries, entry_to_file_map, max_tokens
                )
                file_to_text_map[org_file] = org_content
            except Exception as e:
                logger.error(f"Unable to process file: {org_file}. Skipped indexing it.\nError; {e}", exc_info=True)

        return file_to_text_map, entries, dict(entry_to_file_map)

    @staticmethod
    def process_single_org_file(
        org_content: str,
        org_file: str,
        entries: List[List[Orgnode]],
        entry_to_file_map: List[Tuple[Orgnode, str]],
        max_tokens=256,
        ancestry: Dict[int, str] = {},
    ) -> Tuple[List[List[Orgnode]], List[Tuple[Orgnode, str]]]:
        """Parse org_content from org_file into OrgNode entries

        Recurse down org file entries, one heading level at a time,
        until reach a leaf entry or the current entry tree fits max_tokens.

        Parse recursion terminating entry (trees) into (a list of) OrgNode objects.
        """
        # Prepend the org section's heading ancestry
        ancestry_string = "\n".join([f"{'*' * key} {ancestry[key]}" for key in sorted(ancestry.keys())])
        org_content_with_ancestry = f"{ancestry_string}{org_content}"

        # If content is small or content has no children headings, save it as a single entry
        # Note: This is the terminating condition for this recursive function
        if len(TextToEntries.tokenizer(org_content_with_ancestry)) <= max_tokens or not re.search(
            rf"^\*{{{len(ancestry)+1},}}\s", org_content, re.MULTILINE
        ):
            orgnode_content_with_ancestry = orgnode.makelist(org_content_with_ancestry, org_file)
            entry_to_file_map += zip(orgnode_content_with_ancestry, [org_file] * len(orgnode_content_with_ancestry))
            entries.extend([orgnode_content_with_ancestry])
            return entries, entry_to_file_map

        # Split this entry tree into sections by the next heading level in it
        # Increment heading level until able to split entry into sections or reach max heading level
        # A successful split will result in at least 2 sections
        max_heading_level = 100
        next_heading_level = len(ancestry)
        sections: List[str] = []
        while len(sections) < 2 and next_heading_level < max_heading_level:
            next_heading_level += 1
            sections = re.split(rf"(\n|^)(?=[*]{{{next_heading_level}}} .+\n?)", org_content, flags=re.MULTILINE)

        # If unable to split entry into sections, log error and skip indexing it
        if next_heading_level == max_heading_level:
            logger.error(f"Unable to split current entry chunk: {org_content_with_ancestry[:20]}. Skip indexing it.")
            return entries, entry_to_file_map

        # Recurse down each non-empty section after parsing its body, heading and ancestry
        for section in sections:
            # Skip empty sections
            if section.strip() == "":
                continue

            # Extract the section body and (when present) the heading
            current_ancestry = ancestry.copy()
            first_non_empty_line = [line for line in section.split("\n") if line.strip() != ""][0]
            # If first non-empty line is a heading with expected heading level
            if re.search(rf"^\*{{{next_heading_level}}}\s", first_non_empty_line):
                # Extract the section body without the heading
                current_section_body = "\n".join(section.split(first_non_empty_line, 1)[1:])
                # Parse the section heading into current section ancestry
                current_section_title = first_non_empty_line[next_heading_level:].strip()
                current_ancestry[next_heading_level] = current_section_title
            # Else process the section as just body text
            else:
                current_section_body = section

            # Recurse down children of the current entry
            OrgToEntries.process_single_org_file(
                current_section_body,
                org_file,
                entries,
                entry_to_file_map,
                max_tokens,
                current_ancestry,
            )

        return entries, entry_to_file_map

    @staticmethod
    def convert_org_nodes_to_entries(
        parsed_entries: List[List[Orgnode]],
        entry_to_file_map: Dict[Orgnode, str],
        index_heading_entries: bool = False,
    ) -> List[Entry]:
        """
        Convert OrgNode lists into list of Entry objects

        Each list of OrgNodes is a parsed parent org tree or leaf node.
        Convert each list of these OrgNodes into a single Entry.
        """
        entries: List[Entry] = []
        for entry_group in parsed_entries:
            entry_heading, entry_compiled, entry_raw = "", "", ""
            for parsed_entry in entry_group:
                if not parsed_entry.hasBody and not index_heading_entries:
                    # Ignore title notes i.e notes with just headings and empty body
                    continue

                todo_str = f"{parsed_entry.todo} " if parsed_entry.todo else ""

                # Set base level to current org-node tree's root heading level
                if not entry_heading and parsed_entry.level > 0:
                    base_level = parsed_entry.level
                # Indent entry by 1 heading level as ancestry is prepended as top level heading
                heading = f"{'*' * (parsed_entry.level-base_level+2)} {todo_str}" if parsed_entry.level > 0 else ""
                if parsed_entry.heading:
                    heading += f"{parsed_entry.heading}."

                # Prepend ancestor headings, filename as top heading to root parent entry for context
                # Children nodes do not need ancestors trail as root parent node will have it
                if not entry_heading:
                    ancestors_trail = " / ".join(parsed_entry.ancestors) or Path(entry_to_file_map[parsed_entry])
                    heading = f"* {ancestors_trail}\n{heading}" if heading else f"* {ancestors_trail}."

                compiled = heading

                if parsed_entry.tags:
                    tags_str = " ".join(parsed_entry.tags)
                    compiled += f"\t {tags_str}."

                if parsed_entry.closed:
                    compiled += f'\n Closed on {parsed_entry.closed.strftime("%Y-%m-%d")}.'

                if parsed_entry.scheduled:
                    compiled += f'\n Scheduled for {parsed_entry.scheduled.strftime("%Y-%m-%d")}.'

                if parsed_entry.hasBody:
                    compiled += f"\n {parsed_entry.body}"

                # Add the sub-entry contents to the entry
                entry_compiled += f"{compiled}"
                entry_raw += f"{parsed_entry}"
                if not entry_heading:
                    entry_heading = heading

            if entry_compiled:
                entries.append(
                    Entry(
                        compiled=entry_compiled,
                        raw=entry_raw,
                        heading=f"{entry_heading}",
                        file=f"{entry_to_file_map[parsed_entry]}",
                    )
                )

        return entries



from pathlib import Path

app_root_directory = Path(__file__).parent.parent.parent
web_directory = app_root_directory / "khoj/interface/web/"
next_js_directory = app_root_directory / "khoj/interface/built/"
pypi_static_directory = app_root_directory / "khoj/interface/compiled/"
empty_escape_sequences = "\n|\r|\t| "
app_env_filepath = "~/.khoj/env"
telemetry_server = "https://khoj.beta.haletic.com/v1/telemetry"
content_directory = "~/.khoj/content/"
default_offline_chat_models = [
    "bartowski/Meta-Llama-3.1-8B-Instruct-GGUF",
    "bartowski/Llama-3.2-3B-Instruct-GGUF",
    "bartowski/gemma-2-9b-it-GGUF",
    "bartowski/gemma-2-2b-it-GGUF",
    "Qwen/Qwen2.5-14B-Instruct-GGUF",
]
default_openai_chat_models = ["gpt-4o-mini", "gpt-4o"]
default_gemini_chat_models = ["gemini-1.5-flash", "gemini-1.5-pro"]
default_anthropic_chat_models = ["claude-3-5-sonnet-20241022", "claude-3-5-haiku-20241022"]

empty_config = {
    "search-type": {
        "image": {"encoder": "sentence-transformers/clip-ViT-B-32", "model_directory": "~/.khoj/search/image/"},
    },
}

# default app config to use
default_config = {
    "search-type": {
        "image": {"encoder": "sentence-transformers/clip-ViT-B-32", "model_directory": "~/.khoj/search/image/"},
    },
}



import csv
import json

from apscheduler.job import Job
from django.contrib import admin, messages
from django.contrib.auth.admin import UserAdmin
from django.http import HttpResponse
from django_apscheduler.admin import DjangoJobAdmin
from django_apscheduler.jobstores import DjangoJobStore
from django_apscheduler.models import DjangoJob

from khoj.database.models import (
    Agent,
    ChatModelOptions,
    ClientApplication,
    Conversation,
    Entry,
    GithubConfig,
    KhojUser,
    NotionConfig,
    OpenAIProcessorConversationConfig,
    ProcessLock,
    ReflectiveQuestion,
    SearchModelConfig,
    ServerChatSettings,
    SpeechToTextModelOptions,
    Subscription,
    TextToImageModelConfig,
    UserConversationConfig,
    UserRequests,
    UserVoiceModelConfig,
    VoiceModelOption,
    WebScraper,
)
from khoj.utils.helpers import ImageIntentType

admin.site.unregister(DjangoJob)


class KhojDjangoJobAdmin(DjangoJobAdmin):
    list_display = (
        "id",
        "next_run_time",
        "job_info",
    )
    search_fields = ("id", "next_run_time")
    ordering = ("-next_run_time",)
    job_store = DjangoJobStore()

    def job_info(self, obj):
        job: Job = self.job_store.lookup_job(obj.id)
        return f"{job.func_ref} {job.args} {job.kwargs}" if job else "None"

    job_info.short_description = "Job Info"  # type: ignore

    def get_search_results(self, request, queryset, search_term):
        queryset, use_distinct = super().get_search_results(request, queryset, search_term)
        if search_term:
            jobs = [job.id for job in self.job_store.get_all_jobs() if search_term in str(job)]
            queryset |= self.model.objects.filter(id__in=jobs)
        return queryset, use_distinct


admin.site.register(DjangoJob, KhojDjangoJobAdmin)


class KhojUserAdmin(UserAdmin):
    list_display = (
        "id",
        "email",
        "username",
        "phone_number",
        "is_active",
        "uuid",
        "is_staff",
        "is_superuser",
    )
    search_fields = ("email", "username", "phone_number", "uuid")
    filter_horizontal = ("groups", "user_permissions")

    fieldsets = (
        (
            "Personal info",
            {"fields": ("phone_number", "email_verification_code", "verified_phone_number", "verified_email")},
        ),
    ) + UserAdmin.fieldsets

    actions = ["get_email_login_url"]

    def get_email_login_url(self, request, queryset):
        for user in queryset:
            if user.email:
                host = request.get_host()
                unique_id = user.email_verification_code
                login_url = f"{host}/auth/magic?code={unique_id}"
                messages.info(request, f"Email login URL for {user.email}: {login_url}")

    get_email_login_url.short_description = "Get email login URL"  # type: ignore


admin.site.register(KhojUser, KhojUserAdmin)

admin.site.register(ProcessLock)
admin.site.register(SpeechToTextModelOptions)
admin.site.register(ReflectiveQuestion)
admin.site.register(ClientApplication)
admin.site.register(GithubConfig)
admin.site.register(NotionConfig)
admin.site.register(UserVoiceModelConfig)
admin.site.register(VoiceModelOption)
admin.site.register(UserRequests)


@admin.register(Agent)
class AgentAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
    )
    search_fields = ("id", "name")
    ordering = ("-created_at",)


@admin.register(Entry)
class EntryAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "created_at",
        "updated_at",
        "user",
        "agent",
        "file_source",
        "file_type",
        "file_name",
        "file_path",
    )
    search_fields = ("id", "user__email", "user__username", "file_path")
    list_filter = (
        "file_type",
        "user__email",
        "search_model__name",
    )
    ordering = ("-created_at",)


@admin.register(Subscription)
class KhojUserSubscription(admin.ModelAdmin):
    list_display = (
        "id",
        "user",
        "type",
    )

    search_fields = ("id", "user__email", "user__username", "type")
    list_filter = ("type",)


@admin.register(ChatModelOptions)
class ChatModelOptionsAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "chat_model",
        "model_type",
        "max_prompt_size",
    )
    search_fields = ("id", "chat_model", "model_type")


@admin.register(TextToImageModelConfig)
class TextToImageModelOptionsAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "model_name",
        "model_type",
    )
    search_fields = ("id", "model_name", "model_type")


@admin.register(OpenAIProcessorConversationConfig)
class OpenAIProcessorConversationConfigAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "api_key",
        "api_base_url",
    )
    search_fields = ("id", "name", "api_key", "api_base_url")


@admin.register(SearchModelConfig)
class SearchModelConfigAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "bi_encoder",
        "cross_encoder",
    )
    search_fields = ("id", "name", "bi_encoder", "cross_encoder")


@admin.register(ServerChatSettings)
class ServerChatSettingsAdmin(admin.ModelAdmin):
    list_display = (
        "chat_default",
        "chat_advanced",
        "web_scraper",
    )


@admin.register(WebScraper)
class WebScraperAdmin(admin.ModelAdmin):
    list_display = (
        "priority",
        "name",
        "type",
        "api_key",
        "api_url",
        "created_at",
    )
    search_fields = ("name", "api_key", "api_url", "type")
    ordering = ("priority",)


@admin.register(Conversation)
class ConversationAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "user",
        "created_at",
        "updated_at",
        "client",
    )
    search_fields = ("id", "user__email", "user__username", "client__name")
    list_filter = ("agent",)
    ordering = ("-created_at",)

    actions = ["export_selected_objects", "export_selected_minimal_objects"]

    def export_selected_objects(self, request, queryset):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="conversations.csv"'

        writer = csv.writer(response)
        writer.writerow(["id", "user", "created_at", "updated_at", "conversation_log"])

        for conversation in queryset:
            modified_log = conversation.conversation_log
            chat_log = modified_log.get("chat", [])
            for idx, log in enumerate(chat_log):
                if (
                    log["by"] == "khoj"
                    and log["intent"]
                    and log["intent"]["type"]
                    and (
                        log["intent"]["type"] == ImageIntentType.TEXT_TO_IMAGE.value
                        or log["intent"]["type"] == ImageIntentType.TEXT_TO_IMAGE_V3.value
                    )
                ):
                    log["message"] = "inline image redacted for space"
                    chat_log[idx] = log
            modified_log["chat"] = chat_log

            writer.writerow(
                [
                    conversation.id,
                    conversation.user,
                    conversation.created_at,
                    conversation.updated_at,
                    json.dumps(modified_log),
                ]
            )

        return response

    export_selected_objects.short_description = "Export selected conversations"  # type: ignore

    def export_selected_minimal_objects(self, request, queryset):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="conversations.csv"'

        writer = csv.writer(response)
        writer.writerow(["id", "user", "created_at", "updated_at", "conversation_log"])

        fields_to_keep = set(["message", "by", "created"])

        for conversation in queryset:
            return_log = dict()
            chat_log = conversation.conversation_log.get("chat", [])
            for idx, log in enumerate(chat_log):
                updated_log = {}
                for key in fields_to_keep:
                    updated_log[key] = log[key]
                if (
                    log["by"] == "khoj"
                    and log["intent"]
                    and log["intent"]["type"]
                    and (
                        log["intent"]["type"] == ImageIntentType.TEXT_TO_IMAGE.value
                        or log["intent"]["type"] == ImageIntentType.TEXT_TO_IMAGE_V3.value
                    )
                ):
                    updated_log["message"] = "inline image redacted for space"
                chat_log[idx] = updated_log
            return_log["chat"] = chat_log

            writer.writerow(
                [
                    conversation.id,
                    conversation.user,
                    conversation.created_at,
                    conversation.updated_at,
                    json.dumps(return_log),
                ]
            )

        return response

    export_selected_minimal_objects.short_description = "Export selected conversations (minimal)"  # type: ignore

    def get_actions(self, request):
        actions = super().get_actions(request)
        if not request.user.is_superuser:
            if "export_selected_objects" in actions:
                del actions["export_selected_objects"]
            if "export_selected_minimal_objects" in actions:
                del actions["export_selected_minimal_objects"]
        return actions


@admin.register(UserConversationConfig)
class UserConversationConfigAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "get_user_email",
        "get_chat_model",
        "get_subscription_type",
    )
    search_fields = ("id", "user__email", "setting__chat_model", "user__subscription__type")
    ordering = ("-updated_at",)

    def get_user_email(self, obj):
        return obj.user.email

    get_user_email.short_description = "User Email"  # type: ignore
    get_user_email.admin_order_field = "user__email"  # type: ignore

    def get_chat_model(self, obj):
        return obj.setting.chat_model if obj.setting else None

    get_chat_model.short_description = "Chat Model"  # type: ignore
    get_chat_model.admin_order_field = "setting__chat_model"  # type: ignore

    def get_subscription_type(self, obj):
        if hasattr(obj.user, "subscription"):
            return obj.user.subscription.type
        return None

    get_subscription_type.short_description = "Subscription Type"  # type: ignore
    get_subscription_type.admin_order_field = "user__subscription__type"  # type: ignore



# Generated by Django 4.2.10 on 2024-04-23 17:35

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0035_processlock"),
    ]

    operations = [
        migrations.DeleteModel(
            name="OfflineChatProcessorConversationConfig",
        ),
    ]



# Generated by Django 4.2.5 on 2023-10-18 16:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0007_add_conversation"),
    ]

    operations = [
        migrations.AlterField(
            model_name="conversation",
            name="conversation_log",
            field=models.JSONField(default=dict),
        ),
    ]
